!-----------------------------------------------------------------------------
!
! Read in, tabulate, and interpolate the non-linear power spectra.
! The input file is in the following format (same as CAMB format):
!
! [1st column] k/h (i.e., wavenumber in units of h Mpc^-1) 
! [2nd column] h^3 P11(k) (i.e., P11(k) in units of h^-3 Mpc^3)
! [3rd column] h^3 P22(k) (i.e., P22(k) in units of h^-3 Mpc^3)
! [4th column] h^3 P13(k) (i.e., P13(k) in units of h^-3 Mpc^3)
!
! September 1, 2008
! E. Komatsu
!
!-----------------------------------------------------------------------------
MODULE NonLinearPk
  IMPLICIT none
  INTEGER :: jlo,ndata
  DOUBLE PRECISION, dimension(:), allocatable :: xa1,ya1,y2a1
  DOUBLE PRECISION, dimension(:), allocatable :: xa2,ya2,y2a2
  DOUBLE PRECISION, dimension(:), allocatable :: xa3,ya3,y2a3
contains
  SUBROUTINE Open_NonLinearPkD(filename,n,D)
    IMPLICIT none
    DOUBLE PRECISION, dimension(:), allocatable :: aknl,pk11,pk22,pk13,pk_nl
    CHARACTER(len=128) :: filename
    INTEGER, intent(IN) :: n
    DOUBLE PRECISION, intent(IN) :: D
    INTEGER :: i
    ndata=n
    ALLOCATE(xa1(ndata),ya1(ndata),y2a1(ndata))
    ALLOCATE(aknl(ndata),pk11(ndata),pk22(ndata),pk13(ndata),pk_nl(ndata))
    open(1,file=filename,status='old',form='formatted')
    print*,'read in '//trim(filename)
    do i=1,ndata
       read(1,*)aknl(i),pk11(i),pk22(i),pk13(i) ! k/h & h^3 P(k) in CAMB format!
    enddo
    close(1)
    pk_nl=D**2d0*pk11+D**4d0*(pk22+2d0*pk13)
    xa1=dlog(aknl)
    ya1=dlog(pk_nl)
    DEALLOCATE(aknl,pk_nl,pk11,pk22,pk13)
    CALL spline(xa1,ya1,ndata,1.d30,1.d30,y2a1)
    return
  END SUBROUTINE Open_NonLinearPkD
  SUBROUTINE Close_NonLinearPkD
    DEALLOCATE(xa1,ya1,y2a1)
    return
  END SUBROUTINE Close_NonLinearPkD
  SUBROUTINE Open_NonLinearPkT(filename,n,D)
    IMPLICIT none
    DOUBLE PRECISION, dimension(:), allocatable :: aknl,pk11,pk22,pk13,pk_nl
    CHARACTER(len=128) :: filename
    INTEGER, intent(IN) :: n
    DOUBLE PRECISION, intent(IN) :: D
    INTEGER :: i
    ndata=n
    ALLOCATE(xa2(ndata),ya2(ndata),y2a2(ndata))
    ALLOCATE(aknl(ndata),pk11(ndata),pk22(ndata),pk13(ndata),pk_nl(ndata))
    open(1,file=filename,status='old',form='formatted')
    print*,'read in '//trim(filename)
    do i=1,ndata
       read(1,*)aknl(i),pk11(i),pk22(i),pk13(i) ! k/h & h^3 P(k) in CAMB format!
    enddo
    close(1)
    pk_nl=D**2d0*pk11+D**4d0*(pk22+2d0*pk13)
    xa2=dlog(aknl)
    ya2=dlog(pk_nl)
    DEALLOCATE(aknl,pk_nl,pk11,pk22,pk13)
    CALL spline(xa2,ya2,ndata,1.d30,1.d30,y2a2)
    return
  END SUBROUTINE Open_NonLinearPkT
  SUBROUTINE Close_NonLinearPkT
    DEALLOCATE(xa2,ya2,y2a2)
    return
  END SUBROUTINE Close_NonLinearPkT
  SUBROUTINE Open_NonLinearPkDT(filename,n,D)
    IMPLICIT none
    DOUBLE PRECISION, dimension(:), allocatable :: aknl,pk11,pk22,pk13,pk_nl
    CHARACTER(len=128) :: filename
    INTEGER, intent(IN) :: n
    DOUBLE PRECISION, intent(IN) :: D
    INTEGER :: i
    ndata=n
    ALLOCATE(xa3(ndata),ya3(ndata),y2a3(ndata))
    ALLOCATE(aknl(ndata),pk11(ndata),pk22(ndata),pk13(ndata),pk_nl(ndata))
    open(1,file=filename,status='old',form='formatted')
    print*,'read in '//trim(filename)
    do i=1,ndata
       read(1,*)aknl(i),pk11(i),pk22(i),pk13(i) ! k/h & h^3 P(k) in CAMB format!
    enddo
    close(1)
    pk_nl=D**2d0*pk11+D**4d0*(pk22+2d0*pk13)
    xa3=dlog(aknl)
    ya3=dlog(pk_nl)
    DEALLOCATE(aknl,pk_nl,pk11,pk22,pk13)
    CALL spline(xa3,ya3,ndata,1.d30,1.d30,y2a3)
    return
  END SUBROUTINE Open_NonLinearPkDT
  SUBROUTINE Close_NonLinearPkDT
    DEALLOCATE(xa3,ya3,y2a3)
    return
  END SUBROUTINE Close_NonLinearPkDT
END MODULE NonLinearPk
DOUBLE PRECISION FUNCTION NonLinear_PkD(ak)
  Use NonLinearPk
  IMPLICIT none
  DOUBLE PRECISION :: a,b,h,x,y,ak
  x  = dlog(ak)
  CALL hunt(xa1,ndata,x,jlo)
  h=xa1(jlo+1)-xa1(jlo)
  a=(xa1(jlo+1)-x)/h
  b=(x-xa1(jlo))/h
  y=a*ya1(jlo)+b*ya1(jlo+1)+((a**3-a)*y2a1(jlo)+(b**3-b)*y2a1(jlo+1))*(h**2)/6.
  NonLinear_PkD = dexp(y)
  return
END FUNCTION NonLinear_PkD
DOUBLE PRECISION FUNCTION dlnPkDdlnk(ak)
  Use NonLinearPk
  IMPLICIT none
  DOUBLE PRECISION :: a,b,h,x,y,ak
  x  = dlog(ak)
  CALL hunt(xa1,ndata,x,jlo)
  h=xa1(jlo+1)-xa1(jlo)
  a=(xa1(jlo+1)-x)/h
  b=(x-xa1(jlo))/h
  y=(ya1(jlo+1)-ya1(jlo))/h+(-(3.*a**2-1.)*y2a1(jlo)+(3.*b**2-1.)*y2a1(jlo+1))*h/6.
  dlnPkDdlnk = y
  return
END FUNCTION DlnPkDdlnk
DOUBLE PRECISION FUNCTION NonLinear_PkT(ak)
  Use NonLinearPk
  IMPLICIT none
  DOUBLE PRECISION :: a,b,h,x,y,ak
  x  = dlog(ak)
  CALL hunt(xa2,ndata,x,jlo)
  h=xa2(jlo+1)-xa2(jlo)
  a=(xa2(jlo+1)-x)/h
  b=(x-xa2(jlo))/h
  y=a*ya2(jlo)+b*ya2(jlo+1)+((a**3-a)*y2a2(jlo)+(b**3-b)*y2a2(jlo+1))*(h**2)/6.
  NonLinear_PkT = dexp(y)
  return
END FUNCTION NonLinear_PkT
DOUBLE PRECISION FUNCTION dlnPkTdlnk(ak)
  Use NonLinearPk
  IMPLICIT none
  DOUBLE PRECISION :: a,b,h,x,y,ak
  x  = dlog(ak)
  CALL hunt(xa2,ndata,x,jlo)
  h=xa2(jlo+1)-xa2(jlo)
  a=(xa2(jlo+1)-x)/h
  b=(x-xa2(jlo))/h
  y=(ya2(jlo+1)-ya2(jlo))/h+(-(3.*a**2-1.)*y2a2(jlo)+(3.*b**2-1.)*y2a2(jlo+1))*h/6.
  dlnPkTdlnk = y
  return
END FUNCTION DlnPkTdlnk
DOUBLE PRECISION FUNCTION NonLinear_PkDT(ak)
  Use NonLinearPk
  IMPLICIT none
  DOUBLE PRECISION :: a,b,h,x,y,ak
  x  = dlog(ak)
  CALL hunt(xa3,ndata,x,jlo)
  h=xa3(jlo+1)-xa3(jlo)
  a=(xa3(jlo+1)-x)/h
  b=(x-xa3(jlo))/h
  y=a*ya3(jlo)+b*ya3(jlo+1)+((a**3-a)*y2a3(jlo)+(b**3-b)*y2a3(jlo+1))*(h**2)/6.
  NonLinear_PkDT = dexp(y)
  return
END FUNCTION NonLinear_PkDT
DOUBLE PRECISION FUNCTION dlnPkDTdlnk(ak)
  Use NonLinearPk
  IMPLICIT none
  DOUBLE PRECISION :: a,b,h,x,y,ak
  x  = dlog(ak)
  CALL hunt(xa3,ndata,x,jlo)
  h=xa3(jlo+1)-xa3(jlo)
  a=(xa3(jlo+1)-x)/h
  b=(x-xa3(jlo))/h
  y=(ya3(jlo+1)-ya3(jlo))/h+(-(3.*a**2-1.)*y2a3(jlo)+(3.*b**2-1.)*y2a3(jlo+1))*h/6.
  dlnPkDTdlnk = y
  return
END FUNCTION DlnPkDTdlnk
